// Drawing.java
// A class for rendering structured graphics into a Graphics object
import java.awt.*;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;

public class Drawing
{
	Shape head;
	public double scale = 1.0;
	public int xoffset = 0, yoffset = 0;
	Hashtable known = new Hashtable();

	// constructor
	// Read and parse data from the given stream. Throws an exception
	// on a parse error.
	Drawing(InputStream is) throws IOException
	{
	// Register the default shapes
	register("Line");
	register("Circle");
	register("Text");
	register("Rect");
	register("Poly");

	// Read shapes from input
	StreamTokenizer istk = new StreamTokenizer(is);
	istk.eolIsSignificant(false);
	istk.commentChar('#');
	istk.quoteChar('\"');
	try {
		while(readshape(istk))
			;
		}
	catch(IOException e) {
		throw new IOException(e.getMessage() + " at line " +
				      String.valueOf(istk.lineno()));
		}

	// reverse list of shapes
	Shape ptr, nxt, oldhead=head;
	head = null;
	for(ptr=oldhead; ptr != null; ptr=nxt) {
		nxt = ptr.next;
		ptr.next = head;
		head = ptr;
		}
	}

	// render
	// Render this drawing into a Graphics, with the current zoom and offset
	public void render(Graphics g)
	{
	Shape sh;
	for(sh=head; sh!=null; sh=sh.next)
		sh.render(g, scale, xoffset, yoffset);
	}

	// bounding
	// Returns the bounding box around this drawing, or null
	public Rectangle bounding(Graphics g)
	{
	Rectangle bb = null;
	Shape sh;
	for(sh=head; sh!=null; sh=sh.next) {
		if (bb == null) bb = sh.bounding(g, scale, xoffset, yoffset);
		else bb.add(sh.bounding(g, scale, xoffset, yoffset));
		}
	return bb;
	}

	// register
	// Add a named class to the table of shapes understood by this drawing
	public void register(String cl)
	{
	StringBuffer lower = new StringBuffer(cl);
	lower.setCharAt(0,Character.toLowerCase(lower.charAt(0)));
	try { known.put(lower.toString(), Class.forName(cl)); }
	catch(ClassNotFoundException e) {
		System.out.println("Class "+cl+" does not exist!");
		}
	}

	// readshape
	// Read one shape (line, circle, etc..), adding it to the shapes list.
	boolean readshape(StreamTokenizer istk) throws IOException
	{
	if (istk.nextToken() != StreamTokenizer.TT_WORD) {
		if (istk.ttype == StreamTokenizer.TT_EOF)
			return false;
		else
			throw new IOException("Expected shape name");
		}

	// After reading a shape name, check if a corresponding shape class
	// exists, and if so create one.
	String name = istk.sval;
	Class cl = (Class)known.get(name);
	if (cl == null)
		throw new IOException("Unknown shape name - "+istk.sval);
	Shape sh = null;
	try { sh = (Shape)cl.newInstance(); }
	catch(InstantiationException e) {
		throw new IOException("Failed to instantiate class for shape "+
				      name);
		}
	catch(IllegalAccessException e) {
		throw new IOException("Could not access class for shape "+name);
		}

	// Read shape attribs, passing them to the attrib() function for
	// the newly created shape.
	if (istk.nextToken() != '{')
		throw new IOException("Expected {");
	while(istk.nextToken() != '}') {
		if (!sh.attrib(istk.sval, istk)) 
			throw new IOException("Unknown "+name+" attrib "+
					      istk.sval);
		}
	if (!sh.validate())
		throw new IOException("Missing attributes in "+name);

	// add to shapes list
	sh.next = head;
	head = sh;
	return true;
	}
}


// DPoint
// A simple double-precision point 
class DPoint
{
	double x, y;
	DPoint(double xx, double yy) { x = xx; y = yy; }
	DPoint() { }
}

// Shape
// A simple base class for all the other shapes.
// New shapes should implement render() and bounding(), and override
// attrib() and validate() to parse and check new attributes.
abstract class Shape
{
	Shape next;
	Color colour;
	DPoint pos;

	abstract void render(Graphics g, double s, int xo, int yo);
	abstract Rectangle bounding(Graphics g, double s, int xo, int yo);

	// attrib
	// Called to read an attribute.
	boolean attrib(String name, StreamTokenizer istk) throws IOException
	{
	if (name.equals("colour") || name.equals("color")) {
		// Read R G B triple
		int r,g,b;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected red value");
		r = (int)istk.nval;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected green value");
		g = (int)istk.nval;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected blue value");
		b = (int)istk.nval;
		colour = new Color(r,g,b);
		return true;
		}
	else if (name.equals("pos")) {
		double x,y;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected x value");
		x = istk.nval;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected y value");
		y = istk.nval;
		pos = new DPoint(x,y);
		return true;
		}
	return false;
	}

	// validate
	// Called after reading the shape, to fill in empty fields.
	// Returns false if the shape was not completely read
	boolean validate()
	{
	if (pos == null)
		pos = new DPoint();
	if (colour == null)
		return false;
	return true;
	}
}

class Line extends Shape
{
	Vector points;

	void render(Graphics g, double s, int xo, int yo)
	{
	g.setColor(colour);
	int i;
	for(i=1; i<points.size(); i++) {
		DPoint p1 = (DPoint)points.elementAt(i-1);
		DPoint p2 = (DPoint)points.elementAt(i);
		g.drawLine((int)(p1.x*s+xo), (int)(p1.y*s+yo),
			   (int)(p2.x*s+xo), (int)(p2.y*s+yo));
		}
	}

	Rectangle bounding(Graphics g, double s, int xo, int yo)
	{
	Rectangle bb = null;
	int i;
	for(i=0; i<points.size(); i++) {
		DPoint p = (DPoint)points.elementAt(i);
		int x = (int)(p.x*s+xo), y = (int)(p.y*s+yo);
		if (bb == null) bb = new Rectangle(x,y,1,1);
		else bb.add(x,y);
		}
	return bb;
	}

	boolean attrib(String name, StreamTokenizer istk) throws IOException
	{
	if (name.equals("points")) {
		points = new Vector();
		if (istk.nextToken() != '{')
			throw new IOException("Expected {");
		while(istk.nextToken() != '}') {
			DPoint pt = new DPoint();
			if (istk.ttype != StreamTokenizer.TT_NUMBER)
				throw new IOException("Expected x coordinate");
			pt.x = istk.nval;
			if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
				throw new IOException("Expected y coordinate");
			pt.y = istk.nval;
			points.addElement(pt);
			}
		return true;
		}
	return super.attrib(name, istk);
	}

	boolean validate()
	{
	if (points == null) return false;
	return super.validate();
	}
}

class Circle extends Shape
{
	double r;
	boolean solid;

	void render(Graphics g, double s, int xo, int yo)
	{
	g.setColor(colour);
	if (solid) g.fillOval((int)((pos.x-r)*s+xo), (int)((pos.y-r)*s+yo),
			      (int)(r*2*s), (int)(r*2*s));
	else g.fillOval((int)((pos.x-r)*s+xo), (int)((pos.y-r)*s+yo),
			(int)(r*2*s), (int)(r*2*s));
	}

	Rectangle bounding(Graphics g, double s, int xo, int yo)
	{
	return new Rectangle((int)((pos.x-r)*s+xo), (int)((pos.y-r)*s+yo),
			     (int)(r*2*s), (int)(r*2*s));
	}

	boolean attrib(String name, StreamTokenizer istk) throws IOException
	{
	if (name.equals("solid")) {
		solid = true;
		return true;
		}
	else if (name.equals("radius")) {
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected radius");
		r = istk.nval;
		return true;
		}
	return super.attrib(name, istk);
	}

	boolean validate()
	{
	if (r == 0.0) return false;
	return super.validate();
	}
}

class Text extends Shape
{
	String font;
	double size;
	String text;

	void render(Graphics g, double s, int xo, int yo)
	{
	g.setColor(colour);
	g.setFont(new Font(font, Font.PLAIN, (int)(size*s)));
	g.drawString(text, (int)(pos.x*s+xo), (int)(pos.y*s+yo));
	}

	Rectangle bounding(Graphics g, double s, int xo, int yo)
	{
	g.setFont(new Font(font, Font.PLAIN, (int)(size*s)));
	FontMetrics fnm = g.getFontMetrics();
	return new Rectangle((int)(pos.x*s+xo),
			     (int)(pos.y*s+yo-fnm.getHeight()),
			     (int)(fnm.stringWidth(text)),
			     (int)(fnm.getHeight()));
	}

	boolean attrib(String name, StreamTokenizer istk) throws IOException
	{
	if (name.equals("font")) {
		if (istk.nextToken() != StreamTokenizer.TT_WORD)
			throw new IOException("Expected font name");
		font = istk.sval;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected font size");
		size = istk.nval;
		return true;
		}
	else if (name.equals("text")) {
		if (istk.nextToken() != '\"')
			throw new IOException("Expected text string");
		text = istk.sval;
		return true;
		}
	return super.attrib(name, istk);
	}

	boolean validate()
	{
	if (font == null || size == 0.0 || text == null) return false;
	return super.validate();
	}
}

class Rect extends Shape
{
	DPoint size;
	boolean solid;

	void render(Graphics g, double s, int xo, int yo)
	{
	g.setColor(colour);
	if (solid) g.fillRect((int)(pos.x*s+xo), (int)(pos.y*s+yo),
			      (int)(size.x*s), (int)(size.y*s));
	else g.fillRect((int)(pos.x*s+xo), (int)(pos.y*s+yo),
			(int)(size.x*s), (int)(size.y*s));
	}

	Rectangle bounding(Graphics g, double s, int xo, int yo)
	{
	return new Rectangle((int)(pos.x*s+xo), (int)(pos.y*s+yo),
			     (int)(size.x*s), (int)(size.y*s));
	}

	boolean attrib(String name, StreamTokenizer istk) throws IOException
	{
	if (name.equals("size")) {
		double x,y;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected x value");
		x = istk.nval;
		if (istk.nextToken() != StreamTokenizer.TT_NUMBER)
			throw new IOException("Expected y value");
		y = istk.nval;
		size = new DPoint(x,y);
		return true;
		}
	else if (name.equals("solid")) {
		solid = true;
		return true;
		}
	return super.attrib(name, istk);
	}

	boolean validate()
	{
	if (size == null) return false;
	return super.validate();
	}
}

class Poly extends Line
{
	boolean solid;

	void render(Graphics g, double s, int xo, int yo)
	{
	Polygon p = new Polygon();
	int i;
	for(i=0; i<points.size(); i++) {
		DPoint dpt = (DPoint)points.elementAt(i);
		p.addPoint((int)(dpt.x*s+xo), (int)(dpt.y*s+yo));
		}
	g.setColor(colour);
	if (solid) g.fillPolygon(p);
	else g.drawPolygon(p);
	}

	Rectangle bounding(Graphics g, double s, int xo, int yo)
	{
	Rectangle bb = null;
	int i;
	for(i=0; i<points.size(); i++) {
		DPoint p = (DPoint)points.elementAt(i);
		int x = (int)(p.x*s+xo), y = (int)(p.y*s+yo);
		if (bb == null) bb = new Rectangle(x,y,1,1);
		else bb.add(x,y);
		}
	return bb;
	}

	boolean attrib(String name, StreamTokenizer istk) throws IOException
	{
	if (name.equals("solid")) {
		solid = true;
		return true;
		}
	return super.attrib(name, istk);
	}

	boolean validate()
	{
	return super.validate();
	}
}

