// JFScomponent.java
// The base component from which all JFS client applets are derived. This
// means that a file browser can load a client program to handle some file
// type, and be sure that it has certain needed methods.
// When a JFScomponent is created for interactive user (such as in an Applet
// or a new frame), the creator should -
//  - call the constructor.
//  - call connect() with an authenicated client. Here the component can
//    do any initial loading needed.
//  - call init(). Here the component should build it's user interface,
//    possibly loading data for the UI from the JFS server.
//  - possibly call load() to load some file.
// If a component is being created only to print or render some data, the
// sequence of calls will look like -
//  - call the constructor
//  - call connect()
//  - call load() with the file to print/render
//  - call print() or render()
import java.net.*;
import java.awt.*;
import java.util.Vector;
import JFSclient;

public abstract class JFScomponent extends Panel
{
	JFSclient client;	// connection to JFS server

	// connect
	// Use the given client (which should be already connected and
	// authenticated) for communication. 
	void connect(JFSclient c)
	{
	client = c;
	}

	// init
	// Build the user interface for this component. This method is
	// NOT going to be called if the component is not for interactive use.
	abstract void init();

	// load
	// Load the given file. It must be of some type that this
	// component can handle
	void load(String file, int ver) throws RequestException
	{
	throw new RequestException("No load method defined for "+
				   this.getClass().getName());
	}

	// render
	// Render a representation of the loaded data into a Graphics. This
	// could be called by a program like a word processor that can
	// embed other data into a document.
	void render(Graphics g, int w, int h) { }

	// wantedsize
	// Return the size that this component would like to be. There is
	// no guarantee that it will actually get to be this size though.
	abstract Dimension wantedsize();

	// print
	// Prints the current document to the given printer, which is of
	// the given type (ie. Postscript, Text, ...)
	void print(String printer, String type) throws RequestException
	{
	throw new RequestException("No print method defined for "+
				   this.getClass().getName());
	}

	// windows
	// A vector of all windows associated with this component, to be
	// destroyed when the component is closed. Windows are registered
	// with the reg() method, and unregistered with unreg().
	private Vector windows = new Vector();

	void reg(Frame f) { windows.addElement(f); }

	void unreg(Frame f) { windows.removeElement(f); }

	// close
	// This method should be called when the component ceases to exist
	// (for example, when the window it is in is destroyed). By default,
	// this destroys all windows registered with this component, though
	// subclasses can override this to do other stuff as well.
	void close()
	{
	for(int i=0; i<windows.size(); i++)
		((Frame)windows.elementAt(i)).dispose();
	windows.removeAllElements();
	}

	// save
	// A convenience function for writing data to a file. If the file
	// doesn't already exist then the version actually used is
	// set to something sensible.
	void save(byte data[], String file, String type, int ver, boolean m)
		throws RequestException
	{
	boolean exists = true;
	try { client.statfile(file); }
	catch(RequestException e) {
		exists = false;
		}
	if (!exists) {
		if (m) ver = 1;
		else ver = 0;
		}
	client.put(file, ver, data, type);
	client.setcurrent(file, type);
	}

	// loadimage
	// A convenience function for loading image data from the JFS server
	// and converting it into an AWT image object. Depending on the
	// data type, call an appropriate ImageLoader class to convert to
	// an Image object.
	Image loadimage(String path, int ver) throws RequestException
	{
	// Get the file type and handler
	String lmap[][] = {	{ "image/ppm", "PPMloader" },
				{ "image/x-portable-pixmap", "PPMloader" }
			  };
	String handler = null;
	JFSfile info = client.statfile(path);
	for(int i=0; i<lmap.length; i++)
		if (info.type.equals(lmap[i][0]))
			handler = lmap[i][1];
	if (handler == null)
		throw new RequestException("No ImageLoader for this file type");

	// Load the image loader class
	ImageLoader loader = null;
	try {
		Class lc = Class.forName(handler);
		loader = (ImageLoader)lc.newInstance();
		}
	catch(Exception e) {
		throw new RequestException("Error loading ImageLoader class "+
					   handler);
		}

	// Decode the data
	loader.input(client.get(path, ver));
	return Toolkit.getDefaultToolkit().createImage(loader);
	}
}

